# Generated by Django 4.1.3 on 2022-11-20 22:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("gdaps", "0002_alter_gdapsplugin_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="gdapsplugin",
            name="visible",
        ),
        migrations.AddField(
            model_name="gdapsplugin",
            name="hidden",
            field=models.BooleanField(default=False),
        ),
    ]
